import org.apache.commons.lang3.ArrayUtils
import org.apache.commons.lang3.math.NumberUtils
import org.joda.time.DateTimeConstants

tickUnitCalculator = new TickUnitCalculator()

/**
 * Calculates optimal tick units for JFreeChart charts.
 */
class TickUnitCalculator {
	private static final int MINIMUM_TICK_UNIT_COUNT = 4
	private static final int MAXIMUM_TICK_UNIT_COUNT = 14
	private static final int MAXIMUM_FRACTION = 2
	private static final double OVERALL_ROUNDING_TOLERANCE = 0.2

	/**
	 * Returns the optimal tick unit based on the max time value on that axis.
	 * @param maxValue is the greatest time value to calculate tick size for (measured in seconds).
	 */
	static double calculateTime(double maxValue) {
		if (maxValue == 0) {
			return TimeScaleUnit.HOUR.getSeconds()
		}
		// time based: find the tick that is closest to a round day/hour/minute or its fractions (half, quarter)
		final TimeScaleUnit scaleUnit = getTimeScaleUnit(maxValue)
		long base = scaleUnit.getSeconds()
		double smallestError = maxValue / MAXIMUM_TICK_UNIT_COUNT // we search for anything better than a simple division
		List<Double> optimalTicks = new ArrayList<>()
		// try three bases: original, half, quarter (e.g. 1 hour, 30 min, 15 min)
		for (int fractionIndex = 0; fractionIndex <= MAXIMUM_FRACTION; fractionIndex++) {
			// walk though all possible valid tick counts and store best matches
			for (int tickCount = MAXIMUM_TICK_UNIT_COUNT; tickCount >= MINIMUM_TICK_UNIT_COUNT; tickCount--) {
				double tick = maxValue / tickCount
				double error = tick % base
				if (error <= smallestError) {
					smallestError = error
					if (error <= OVERALL_ROUNDING_TOLERANCE * tick) {
						optimalTicks.add(tick)
					}
				}
			}
			base /= 2
		}
		double bestTick = optimalTicks.isEmpty() ?
				maxValue / MAXIMUM_TICK_UNIT_COUNT :
				NumberUtils.min(ArrayUtils.toPrimitive(optimalTicks.toArray(new Double[0]))) // smaller tick size is better
		return round(bestTick)
	}

	/**
	 * Rounds a time value to the closest and biggest "round" value (day, half day, quarter day, hour, half hour, etc.)
	 * If the input value is not zero, the result will always be at least rounded UP to the nearest minute.
	 * @param value time duration in seconds
	 */
	static int round(double value) {
		for (TimeScaleUnit unit = TimeScaleUnit.DAY; unit.hasSmallerUnit(); unit = unit.getSmallerUnit()) {
			long base = unit.getSeconds()
			for (int fractionIndex = 0; fractionIndex <= MAXIMUM_FRACTION; fractionIndex++) {
				if (Math.min(value % base, base - (value % base)) <= OVERALL_ROUNDING_TOLERANCE * value) {
					int roundedValue = (int) (Math.round(value / base) * base)
					if (roundedValue > 0) {
						return roundedValue
					}
				}
				base /= 2
			}
		}
		return (int) (Math.ceil(value / TimeScaleUnit.MINUTE.getSeconds()) * TimeScaleUnit.MINUTE.getSeconds()) // always round up to minute
	}

	/**
	 * Returns a reasonable time unit to use for calculating the number of ticks.
	 */
	private static TimeScaleUnit getTimeScaleUnit(double seconds) {
		if (seconds >= MAXIMUM_TICK_UNIT_COUNT * TimeScaleUnit.DAY.getSeconds()) {
			return TimeScaleUnit.DAY
		} else if (seconds >= MAXIMUM_TICK_UNIT_COUNT * TimeScaleUnit.HOUR.getSeconds()) {
			return TimeScaleUnit.HOUR
		}
		return TimeScaleUnit.MINUTE
	}

	public enum TimeScaleUnit {
		MINUTE(DateTimeConstants.SECONDS_PER_MINUTE),
		HOUR(DateTimeConstants.SECONDS_PER_HOUR),
		DAY(DateTimeConstants.SECONDS_PER_DAY)

		private final long seconds

		TimeScaleUnit(long seconds) {
			this.seconds = seconds
		}

		public long getSeconds() {
			return seconds
		}

		public boolean hasSmallerUnit() {
			return this.ordinal() > 0
		}

		public TimeScaleUnit getSmallerUnit() {
			return TimeScaleUnit.values()[this.ordinal() - 1]
		}
	}
}
